use core::time::Duration;
use std::collections::HashMap;

use bitfinex::api::Bitfinex;
use bitfinex::positions::Position;
use bitfinex::ticker::TradingPairTicker;
use futures_util::stream::FuturesUnordered;
use tokio::stream::StreamExt;
use tokio::time::delay_for;

use crate::connectors::{Client, Connector, ExchangeKind};
use crate::currency::{Symbol, SymbolPair};
use crate::events::{Event, EventKind};
use crate::managers::{OrderManager, PositionManager, PriceManager};
use crate::strategy::PositionStrategy;
use crate::ticker::Ticker;
use crate::BoxError;

pub struct BfxBot {
    ticker: Ticker,
    quote: Symbol,
    trading_symbols: Vec<Symbol>,
    price_managers: Vec<PriceManager>,
    order_managers: Vec<OrderManager>,
    pos_managers: Vec<PositionManager>,
}

impl BfxBot {
    pub fn new(
        exchanges: Vec<ExchangeKind>,
        trading_symbols: Vec<Symbol>,
        quote: Symbol,
        tick_duration: Duration,
    ) -> Self {
        let clients: Vec<_> = exchanges.iter().map(|x| Client::new(x)).collect();

        let mut pos_managers = Vec::new();
        let mut order_managers = Vec::new();
        let mut pair_statuses = Vec::new();

        for c in clients {
            pos_managers.push(PositionManager::new(c.clone()));
            order_managers.push(OrderManager::new(c.clone()));
            pair_statuses.extend(
                trading_symbols
                    .iter()
                    .map(|x| SymbolPair::new(quote.clone(), x.clone()))
                    .map(|x| PriceManager::new(x, c.clone()))
                    .collect::<Vec<PriceManager>>(),
            )
        }

        BfxBot {
            ticker: Ticker::new(tick_duration),
            price_managers: pair_statuses,
            quote,
            trading_symbols,
            order_managers,
            pos_managers,
        }
    }

    pub async fn start_loop(&mut self) -> Result<(), BoxError> {
        self.update_managers().await?;

        loop {
            self.update().await;
        }
    }

    async fn update(&mut self) -> Result<(), BoxError> {
        delay_for(self.ticker.duration()).await;
        self.ticker.inc();

        self.update_managers().await?;

        Ok(())
    }

    async fn update_managers(&mut self) -> Result<(), BoxError> {
        self.update_price_managers().await?;

        Ok(())
    }

    async fn update_price_managers(&mut self) -> Result<Option<Vec<Event>>, BoxError> {
        let futures: Vec<_> = self
            .price_managers
            .clone()
            .into_iter()
            // the only reason you need the async block is that the future
            // returned by x.update(tick) borrows from x
            // so we create a future that first takes ownership of x, then uses it to call x.update
            .map(|mut x| {
                let tick = self.ticker.current_tick();
                async move { x.update(tick).await }
            })
            .map(tokio::spawn)
            .collect();

        let mut price_entries = vec![];

        for f in futures {
            price_entries.push(f.await??);
        }

        for mut manager in &mut self.price_managers {
            let prices: Vec<_> = price_entries
                .drain_filter(|x| x.pair() == manager.pair())
                .collect();

            for p in prices {
                manager.add_entry(p);
            }
        }

        Ok(None)
    }
}
